/*   Copyright (c) 1988-1991 Shamus Software Ltd. */


/* One time pad - uses an 'unpredictable' random number generator  *
 * the output of which is XORed with the plaintext. XORing again   *
 * with the output of the same generator at the other end restores *
 * the plaintext. Now uses 'Cipher chaining' for greater security. */

#include <stdio.h>
#include <miracl.h>

/* define LOOPBACK if debugging */

static big compo,xin,xout;
static int rand_words,rand_bytes;
static int in,out;
static char *iptr,*optr;
static bool started=FALSE;
static unsigned char imkey[100],omkey[100];

static void reverse(x)
big x;
{ /* reverse digits of x */
    small t;
    int i,m;
    for (i=1;i<=x[0]/2;i++)
    {
        m=x[0]-i+1;
        t=x[i];
        x[i]=x[m];
        x[m]=t;
    }
}

void setpad(composite,key,mode)
big composite;
big key;
int mode;
{ /* initialise one time pad */
    int i;
    xin=mirvar(0);
    xout=mirvar(0);
    compo=mirvar(0);
    copy(composite,compo);
    normalise(compo,compo);
    copy(key,xin);
    copy(key,xout);
#ifndef LOOPBACK
    if (mode<2) reverse(xin);   /*    use different pads.. */
    else        reverse(xout);  /*    ..for each direction */
#endif
    mad(xin,xin,xin,compo,compo,xin);
    mad(xout,xout,xout,compo,compo,xout);
    rand_words=(compo[0]-1)/2; /* use least significant bytes only */
    rand_bytes=(SBITS/8)*rand_words; 
    for (i=0;i<rand_bytes;i++) omkey[i]=imkey[i]=0;
    in=out=0;

    iptr=(char *)&xin[1];
    optr=(char *)&xout[1];
    started=TRUE;
}

void end_pad()
{ /* finish pad */
    started=FALSE;
    zero(xin);
    zero(xout);
    zero(compo);
    free(xin);
    free(xout);
    free(compo);
}

int encode(ch)
int ch;
{ /* encode ch using one-time pad */
    if (!started) return ch;
    ch^=*optr;
    omkey[out]=ch;
    optr++;
    out++;
    if (out>=rand_bytes)
    { /* run out of bytes - generate some more */
        optr=(char *)&xout[1];
        for (out=0;out<rand_bytes;out++,optr++) *optr^=omkey[out];
        mad(xout,xout,xout,compo,compo,xout);
        optr=(char *)&xout[1];
        out=0;
    }
    return ch;
}

int decode(ch)
int ch;
{ /* decode ch using one-time pad */
    if (!started) return ch;
    imkey[in]=ch;
    ch^=*iptr;
    iptr++;
    in++;
    if (in>=rand_bytes)
    { /* run out of bytes - generate some more */
        iptr=(char *)&xin[1];
        for (in=0;in<rand_bytes;in++,iptr++) *iptr^=imkey[in];
        mad(xin,xin,xin,compo,compo,xin);
        iptr=(char *)&xin[1];
        in=0;
    }
    return ch&0xff;
}

